import json
from pathlib import Path
from sklearn.metrics import confusion_matrix
from sklearn.metrics._plot import confusion_matrix as plt_CM
import matplotlib.pyplot as plt
from tqdm import tqdm

AVAILABLE_CLASSES = ['kicking', 'punching', 'pushing', 'neutral']


def get_model_stat(model_data_path: Path):
    return json.load(model_data_path.open(mode='r'))


def get_accuracy(pred, truth):
    accuracy = 0
    for i in range(len(truth)):
        if pred[i] == truth[i]:
            accuracy += 1
    return accuracy / len(truth)


def get_fp_tp(truth, pred, label):
    fp = 0
    tp = 0
    fn = 0
    for i in range(0, len(truth) - 1):
        if pred[i] == label and truth[i] == label:
            tp += 1
        elif pred[i] != label and truth[i] == label:
            fn += 1
        elif pred[i] == label and truth[i] != label:
            fp += 1
        else:
            a = 1
    return tp, fp, fn


def precision(TP, FP):
    if TP == 0:
        return 0
    else:
        return TP / (TP + FP)


def recall(TP, FN):
    if TP == 0:
        return 0
    else:
        return TP / (TP + FN)


def F1_score(P, R):
    if P == 0 and R == 0:
        return 0
    else:
        return 2 * P * R / (P + R)


def model_visualisation(model_stat: dict):
    """
    Plot the confusion matrix and data about the model
    """
    # confusion matrix
    C_M_train = confusion_matrix(model_stat["y_true_train"],
                                 model_stat["y_pred_train"])
    C_M_test = confusion_matrix(model_stat["y_true_test"],
                                model_stat["y_pred_test"])
    CM_visualizer = plt_CM.ConfusionMatrixDisplay(C_M_train, display_labels=AVAILABLE_CLASSES)
    CM_visualizer.plot()
    plt.title("Training Confusion Matrix")
    plt.show()
    CM_visualizer = plt_CM.ConfusionMatrixDisplay(C_M_test, display_labels=AVAILABLE_CLASSES)
    CM_visualizer.plot()
    plt.title("Testing Confusion Matrix")
    plt.show()
    accuracy = get_accuracy(model_stat["y_true_test"],
                            model_stat["y_pred_test"])
    # data
    print("Accuracy : ", accuracy)
    print(model_stat["classifier_name"])
    print(model_stat["hyperparams"])
    print(model_stat["tracker_params"])
    print(model_stat["augment_grid"])


def evaluate(model_stat: dict, print_flag: bool = False):
    """
    Evaluate one model and print data if flag turned on
    """
    train_acc = get_accuracy(model_stat["y_true_train"], model_stat["y_pred_train"])
    test_acc = get_accuracy(model_stat["y_true_test"], model_stat["y_pred_test"])
    list_f1_score = []
    list_recall = []
    list_precision = []
    for label in range(len(AVAILABLE_CLASSES)):
        tp, fp, fn = get_fp_tp(model_stat["y_true_test"], model_stat["y_pred_test"], label)
        recallval = recall(tp, fn)
        precisionval = precision(tp, fp)
        f1_scoreval = F1_score(precisionval, recallval)
        list_f1_score.append(f1_scoreval)
        list_recall.append(recallval)
        list_precision.append(precisionval)
        if print_flag:
            print(AVAILABLE_CLASSES[label], " F1 score: ", round(f1_scoreval, 3))
    avg_scoreval = sum(list_f1_score) / len(list_f1_score)
    avg_recall = sum(list_recall) / len(list_recall)
    avg_precision = sum(list_precision) / len(list_precision)
    if print_flag:
        print("Average F1 score:", round(avg_scoreval, 3))
        print("Average Precision:", round(avg_precision, 3))
        print("Average Recall:", round(avg_recall, 3))
        print("Training accuracy: ", round(train_acc, 3))
        print("Testing accuracy: ", round(test_acc, 3))
    return avg_scoreval

def whoisbest(model_data_folder_path: Path = Path("data\models\evaluation")):
    """

    Args:
        model_data_folder_path: Path of model .json files

    Returns:
        plot and print performances and hyperparameters of the best model optimizing Macro F1 Score
    """
    best_model = None
    path_best = None

    for i in tqdm(model_data_folder_path.iterdir()):
        model_stat = get_model_stat(i)
        if best_model == None:
            best_model = model_stat
            path_best = i

        if evaluate(model_stat) > evaluate((best_model)):
            path_best = i
            best_model = model_stat
    evaluate(best_model, True)
    print("Best model path : ", path_best)

    model_visualisation(best_model)



"""
keep a rolling window for every skeleton on a frame. If the skeletons
disappear, its rolling window is going to be deleted.

they are all tracked inside a dictionnary that has the tracking id
as a key e.g.
{
    1: SkeletonRollingWindow(...),
    2: SkeletonRollingWindow(...),
    6: SkeletonRollingWindow(...),
    12: SkeletonRollingWindow(...),
}
"""
from typing import Union, Generator, Tuple, List, Dict
from collections import deque, Counter
from collections.abc import Sequence
import numpy as np
from tactus_data import SkeletonRollingWindow, Skeleton


class FeatureTracker:
    """
    High level interface with rolling windows of skeletons and their
    tracking with deepsort.
    """
    def __init__(self,
                 window_size: int = 5,
                 angles_to_compute: List[Tuple[int, int, int]] = None
                 ):
        self.window_size = window_size
        self.angles_to_compute = angles_to_compute

        self.rolling_windows: dict[int, SkeletonRollingWindow]
        self.reset_rolling_windows()

        self.tracks_to_del = []

    def reset_rolling_windows(self):
        self.rolling_windows = {}

    def update_rolling_window(self, track_id: int, skeleton: dict):
        """update a SkeletonRollingWindow from its ID"""
        if track_id not in self.rolling_windows:
            self.rolling_windows[track_id] = SkeletonRollingWindow(self.window_size, self.angles_to_compute)

        self.rolling_windows[track_id].add_skeleton(skeleton)

    def duplicate_last_entry(self, track_id: int, new_bbox_lbrt: Tuple[float, float, float, float]) -> Skeleton:
        """
        duplicate the last entry of a rolling window. That is useful
        to avoid gaps in the prediction.
        """
        return self.rolling_windows[track_id].duplicate_last_entry(new_bbox_lbrt)

    def delete_track_id(self, track_id: int):
        """
        delete a SkeletonRollingWindow from its ID.

        DeepSort can have unconfirmed track that are not taken into
        account in FeatureTracker. To avoid an error, we delete
        the index only if it exists.
        """
        if track_id in self.rolling_windows:
            del self.rolling_windows[track_id]

    def extract(self) -> Generator[Tuple[int, np.ndarray], None, None]:
        """
        Extract features from each SkeletonRollingWindow

        Yields
        ------
        Generator[int, np.ndarray]
            yields (track_id, features) for each SkeletonRollingWindow
        """
        for track_id, rolling_window in self.rolling_windows.items():
            yield track_id, rolling_window.get_features()

    def __getitem__(self, __track_id: int) -> SkeletonRollingWindow:
        return self.rolling_windows[__track_id]


class PredTracker:
    """
    save the non-neutral predictions for each skeleton still present
    on the stream.
    """
    def __init__(self, prediction_smoothing: int = 5):
        self.tracker: Dict[int, Dict] = {}
        self.pred_rw_size = prediction_smoothing

    def init_track(self, track_id: int):
        """initialize the dictionary key"""
        self.tracker[track_id] = {"current_label": None,
                                  "violent": False,
                                  "pred_history": deque(maxlen=self.pred_rw_size)}

    def add_pred(self, track_id: int, label: str):
        """starts the tracking of a person from a prediction label"""
        if track_id not in self.tracker:
            self.init_track(track_id)

        self.tracker[track_id]["pred_history"].append(label)
        smoothed_label = self.get_smoothed_pred(track_id)
        self.tracker[track_id]["current_label"] = smoothed_label

        if smoothed_label != "neutral":
            self.tracker[track_id]["violent"] = True

    def delete_track_id(self, track_ids: Union[List[int], int]):
        """removes the track of a person"""
        if not isinstance(track_ids, Sequence):
            track_ids = [track_ids]

        for track_id in track_ids:
            if track_id in self.tracker:
                del self.tracker[track_id]

    def get_smoothed_pred(self, track_id: int) -> str:
        """get the smoothed prediction"""
        return most_common(self.tracker[track_id]["pred_history"])

    def get_last_pred(self, track_id: int) -> str:
        """return the last prediction made"""
        return self.tracker[track_id]["current_label"]

    def __contains__(self, __track_id: int):
        return __track_id in self.tracker

    def __getitem__(self, __track_id: int):
        return self.tracker[__track_id]


def most_common(lst):
    """return the most common element of a list"""
    counter = Counter(lst)
    return counter.most_common(1)[0][0]



from typing import List, Dict, Tuple, Generator
import json
from pathlib import Path
import numpy as np
from imblearn.over_sampling import SMOTE
from imblearn.under_sampling import RandomUnderSampler
from tactus_data import data_augment, SMALL_ANGLES_LIST, MEDIUM_ANGLES_LIST
from tactus_data.datasets.ut_interaction import data_split
from tactus_data import Skeleton

from tactus_model.utils.tracker import FeatureTracker
from tactus_model.utils.classifier import Classifier, label_to_int

DATA_AUGMENT_GRIDS = {
    "FLIP": {"horizontal_flip": [True, False]},
}

TRACKER_GRID = {
    "window_size": [9],
    "number_of_angles": [0],
}

CLASSIFIER_HYPERPARAMS = {
    "TorchMLP": {
        "batch_size": [256],
        "num_epochs": [30],
        "hidden_layer_sizes": [{"layer_sizes": [256, 128, 16], "dropout": [0.4, 0.2, 0.1]},
                               {"layer_sizes": [1024, 512, 128, 16], "dropout": [0.5, 0.3, 0.2, 0.1]},
                               {"layer_sizes": [512, 128, 64], "dropout": [0.2, 0.2, 0.1]},
                               ],
        "activation": ['ReLU']
    },
}


def get_classifier(
        classifier_grids: Dict[str, Dict] = None,
) -> Generator[Tuple[Classifier, str, dict], None, None]:
    """
    get a classifier instance with a set of hyperparametres found in
    the grid.

    Yields
    ------
    Generator[Tuple[Classifier, str, dict], None, None]
        a tuple with
        - the classifier initialised with the hyperparametres
        - the classifier name
        - the hyperparamtres used on the classifier
    """
    if classifier_grids is None:
        classifier_grids = CLASSIFIER_HYPERPARAMS

    for classifier_name, hyperparams_grid in classifier_grids.items():
        for hyperparams in data_augment.ParameterGrid(hyperparams_grid):
            hyperparams["dropout_layers_sizes"] = hyperparams["hidden_layer_sizes"]["dropout"]
            hyperparams["hidden_layer_sizes"] = hyperparams["hidden_layer_sizes"]["layer_sizes"]
            yield Classifier(classifier_name, hyperparams)


def get_augment_grid(
        augment_grids: Dict[str, Dict] = None,
) -> Generator[dict, None, None]:
    """
    get a data augmentation grid from augment_grids.

    Yields
    ------
    Generator[dict, None, None]
    """
    if augment_grids is None:
        augment_grids = DATA_AUGMENT_GRIDS

    for _, augment_grid in augment_grids.items():
        yield augment_grid


def get_tracker_params(
        tracker_grid: Dict[str, List] = None,
) -> Generator[dict, None, None]:
    """
    get a tracker parametres (with window size and number of angle) from
    the grid.

    Yields
    ------
    Generator[dict, None, None]
    """
    if tracker_grid is None:
        tracker_grid = TRACKER_GRID

    for grid in data_augment.ParameterGrid(tracker_grid):
        yield grid


def sample_data(X: List, Y: List, sampler: str):
    """
    Over/Under sample the data if the flag is set to the corresponding value, does nothing if flag = X
    Args:
        X: Data for the classifier
        Y: Label for the classifier
        sampler: Flag chosing the sample technique

    Returns:
        The sampled data
    """
    if sampler == "RUS":
        sampler = RandomUnderSampler()
    elif sampler == "SMOTE":
        sampler = SMOTE()
    else:
        return X, Y
    return sampler.fit_resample(X, Y)


def train_grid_search(
        fps: int = 10,
        augment_grids: Dict[str, Dict] = None,
        tracker_grid: Dict[str, List] = None,
        classifier_grids: Dict[str, Dict] = None,
        sampler: str = "RUS"
):
    """
    launch the training process

    Parameters
    ----------
    fps : int, optional
        the fps to train on, by default 10
    """
    # cant use a generator here because we use this multiple times
    train_videos, _, test_videos = data_split(Path("data/processed/ut_interaction/"), (85, 0, 15))

    model_id = 0
    for augment_grid in get_augment_grid(augment_grids):
        # augments data and saves it in files
        print("augments data with: ", augment_grid)
        delete_data_augment(train_videos + test_videos, fps)

        for video_path in train_videos:
            original_data_path = video_path / f"{fps}fps" / "yolov7.json"
            data_augment.grid_augment(original_data_path, augment_grid)

        for video_path in test_videos:
            original_data_path = video_path / f"{fps}fps" / "yolov7.json"
            data_augment.grid_augment(original_data_path, {})

        # compute features
        for tracker_params in get_tracker_params(tracker_grid):
            print("compute features with: ", tracker_params)
            angle_list = get_angle_list(tracker_params["number_of_angles"])
            window_size = tracker_params["window_size"]

            X, Y = generate_features(train_videos, fps, window_size, angle_list)
            X_test, Y_test = generate_features(test_videos, fps, window_size, angle_list)
            X,Y = sample_data(X, Y, sampler)

            for classifier in get_classifier(classifier_grids):
                print("fit classifier: ", classifier.name, " with ", classifier.hyperparams)
                classifier.fit(X, Y)
                classifier.window_size = window_size
                classifier.fps = fps
                classifier.angles_to_compute = angle_list
                classifier.save(Path(f"data/models/pickle/{model_id}.pickle"))

                save_model_evaluation(model_id, classifier, augment_grid,
                                      tracker_params, X, Y, X_test, Y_test)

                model_id += 1


def save_model_evaluation(model_id, classifier, augment_grid, tracker_params, X, Y, X_test, Y_test):
    """
    predict the train and the test dataset to measure the performance
    of the model later on.
    """
    eval_dict = {}

    eval_dict["augment_grid"] = augment_grid
    eval_dict["tracker_params"] = tracker_params

    eval_dict["classifier_name"] = classifier.name
    eval_dict["hyperparams"] = classifier.hyperparams

    eval_dict["y_pred_train"] = classifier.predict(X)
    eval_dict["y_true_train"] = Y

    eval_dict["y_pred_test"] = classifier.predict(X_test)
    eval_dict["y_true_test"] = Y_test

    filename = Path(f"data/models/evaluation/{model_id}.json")
    filename.parent.mkdir(parents=True, exist_ok=True)
    json.dump(eval_dict, filename.open(mode="w", encoding="utf8"))


def generate_features(videos: List[Path], fps: int, window_size: int, angle_list: List):
    """
    generates features for a list of video directories. The directories
    must follow the following structure:
    `video_name -> label.json`
    `video_name -> xxfps -> *_augment_*.json`

    Parameters
    ----------
    videos : List[Path]
        list of the video directories.
    fps : int
        the fps to generate features with.
    window_size : int
        the size of the rolling window.
    angle_list : List
        the list of the angles to compute.

    Returns
    -------
    (X, Y)
        return a tuple with the features and the labels.
    """
    X = []
    Y = []

    for video_path in videos:
        label_path = video_path / (f"{video_path.stem}.label.json")
        labels = json.load(label_path.open())

        for augmented_data_path in video_path.glob(f"{fps}fps/*_augment_*.json"):
            augmented_data = json.load(augmented_data_path.open())
            video_features, video_labels = feature_from_video(augmented_data, labels, window_size, angle_list)

            X.extend(video_features)
            Y.extend(video_labels)

    return X, Y


def feature_from_video(
    formatted_json: Dict,
    labels: Dict,
    window_size: int,
    angle_list: List
):
    """
    Compute features and true labels from a `json video` file.

    Parameters
    ----------
    formatted_json : Dict
        video json file containing the list of frames and their
        skeletons.
    labels : Dict
        the content of the video label file.
    window_size : int
        the size of the rolling window.
    angle_list : List
        the list of angles to compute in the features.

    Returns
    -------
    (X, Y)
        return a tuple with the features and the labels.
    """
    feature_tracker = FeatureTracker(window_size=window_size, angles_to_compute=angle_list)

    video_features = []
    video_labels = []

    offender_id = labels["offender"][0]

    i_label = 0
    i_frame = get_i_frame(labels["classes"][i_label]["start_frame"], formatted_json["frames"])

    while i_frame < len(formatted_json["frames"]):
        frame = formatted_json["frames"][i_frame]

        for skeleton in frame["skeletons"]:
            # do not deal with untracked skeletons
            if "id_stupid" in skeleton:
                skeleton_id = skeleton["id_stupid"]

                keypoints = np.array(skeleton["keypoints"]).reshape((-1, 2))
                feature_tracker.update_rolling_window(skeleton_id, Skeleton(keypoints=keypoints))

                if feature_tracker[skeleton_id].is_complete():
                    features = feature_tracker[skeleton_id].get_features()

                    if skeleton_id == offender_id:
                        label = compute_label(frame["frame_id"], labels["classes"], i_label)
                    else:
                        label = "neutral"

                    video_features.append(features)
                    video_labels.append(label_to_int(label))

        # jump to the next action
        if i_label >= labels["classes"][i_label]["end_frame"]:
            if len(labels["classes"]) > i_label:
                i_label += 1
                i_frame = labels["classes"][i_label]["start_frame"] - window_size
                feature_tracker.reset_rolling_windows()
            else:
                break
        else:
            i_frame += 1

    return video_features, video_labels


def compute_label(frame_id: str, classes: List[Dict], i_label: int) -> str:
    """
    compute the label of the offender for a given frame.

    Parameters
    ----------
    frame_id : str
        the frame id. Usually "xxx.jpg".
    classes : list
        list of labels `classes`. It must have `end_frame`,
        `end_frame`, `classification` keys.
    i_label : int
        the index of the current class.

    Returns
    -------
    str
        the corresponding label.
    """
    frame_id = frame_id_to_int(frame_id)

    if classes[i_label]["start_frame"] < frame_id < classes[i_label]["end_frame"]:
        return classes[i_label]["classification"]

    return "neutral"


def get_angle_list(number_of_angles: int) -> List[Tuple[int, int, int]]:
    """
    get predefined list of angles to compute from the number
    of angles we would like.

    Parameters
    ----------
    number_of_angles : int
        number of angles to compute. Must be 0, 4 or 8.

    Returns
    -------
    list[tuple[int, int, int]]
        the list of angles to compute.
    """
    if number_of_angles == 0:
        return []

    if number_of_angles == 4:
        return SMALL_ANGLES_LIST

    if number_of_angles == 8:
        return MEDIUM_ANGLES_LIST


def delete_data_augment(video_paths: List[Path], fps: int):
    """
    delete every augmentation data. Necessary in case the new
    augmentation is smaller than the former one.

    Parameters
    ----------
    video_paths : List[Path]
        list of the video directories.
    fps : int
        the fps of the data augmentation to delete.
    """
    for video_path in video_paths:
        for augmented_data_path in video_path.glob(f"{fps}fps/*_augment_*"):
            augmented_data_path.unlink()


def get_i_frame(target_frame_id: int,
                frames: List):
    """
    convert a frame_id to its index in a list.

    Parameters
    ----------
    starting_frame: int
        number of the starting frame.
    frames: List
        list containing frame info.
    """
    frame_index = 0
    current_frame_id = frame_id_to_int(frames[frame_index]["frame_id"])

    while current_frame_id < target_frame_id:
        current_frame_id = frame_id_to_int(frames[frame_index]["frame_id"])
        frame_index += 1
    return frame_index


def frame_id_to_int(frame_id: str) -> int:
    """
    convert the id of a frame (003.jpg) to an int (3).

    removesuffix is not available in python 3.8
    """
    return int(frame_id[:-4])


if __name__ == "__main__":
    train_grid_search()



from typing import Tuple, List
from tqdm import tqdm
import torch
from torch import nn
from torch.utils.data import DataLoader, TensorDataset
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import _num_features


class TorchMLP:
    def __init__(
        self,
        hidden_layer_sizes: Tuple[int],
        activation: str,
        *,
        dropout_layers_sizes: Tuple[int] = None,
        num_epochs: Tuple[int] = None,
        batch_size: Tuple[int] = None,
        device: str = None,
    ):
        if dropout_layers_sizes is None:
            dropout_layers_sizes = [0] * len(hidden_layer_sizes)
        if len(hidden_layer_sizes) != len(dropout_layers_sizes):
            raise IndexError('dropout_layers must be the same size as hidden_layer_sizes')

        self.hidden_layer_sizes = hidden_layer_sizes
        self.activation = activation
        self.dropout_layers_sizes = dropout_layers_sizes

        self.num_epochs = num_epochs
        self.batch_size = batch_size

        self.model = None
        self.loss_fn = None
        self.optimizer = None

        self.select_device(device)

    def fit(self, X, Y, *, num_epochs: int = None, batch_size: int = None):
        """
        fit the model

        Parameters
        ----------
        num_epochs : int
            number of epochs
        """
        num_epochs = num_epochs or self.num_epochs
        batch_size = batch_size or self.batch_size
        if num_epochs is None:
            raise ValueError("num_epochs should be either defined as an instance attribute "
                             "or in the the fit function")
        if batch_size is None:
            raise ValueError("batch_size should be either defined as an instance attribute "
                             "or in the the fit function")

        input_size = _num_features(X)
        output_size = len(unique_labels(Y))

        self._build(input_size, output_size)
        self.model.train()

        if isinstance(X, List):
            X = torch.tensor(X, dtype=torch.float32, device=self.device)
        if isinstance(Y, List):
            Y = torch.tensor(Y, dtype=torch.long, device=self.device)

        dataset = TensorDataset(X, Y)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

        for _ in tqdm(range(num_epochs)):
            for _, (x_batch, y_batch) in enumerate(dataloader):
                y_pred = self.model(x_batch)
                loss = self.loss_fn(y_pred, y_batch)
                self.optimizer.zero_grad()
                loss.backward()
                self. optimizer.step()

        torch.cuda.empty_cache()

    def predict(self, X):
        """
        predict using the fitted model
        """
        self.model.eval()
        if isinstance(X, List):
            X = torch.tensor(X, dtype=torch.float32, device=self.device)

        return torch.argmax(self.model(X), dim=1)

    def _build(self, input_size: int, output_size: int):
        """
        automatically build the model in regard of the input and the
        output size.

        Parameters
        ----------
        input_size : int
            size of a feature
        output_size : int
            number of different classes
        """
        layer_sizes = [input_size]
        layer_sizes.extend(self.hidden_layer_sizes)
        layer_sizes.append(output_size)

        layers = []
        for i in range(len(layer_sizes) - 1):
            layers.append(nn.Linear(layer_sizes[i], layer_sizes[i + 1]))
            if i < len(layer_sizes) - 2:
                activation_layer = getattr(nn, self.activation)
                layers.append(activation_layer())
                layers.append(nn.Dropout(self.dropout_layers_sizes[i]))

        layers.append(nn.Softmax(dim=1))
        self.model = nn.Sequential(*layers).to(device=self.device)

        self.loss_fn = nn.CrossEntropyLoss()
        self.optimizer = torch.optim.Adam(self.model.parameters())

    def select_device(self, device: str = None):
        """
        select the device or default to CPU if not found.

        Parameters
        ----------
        device : str, optional
            name of the device to use, by default None
        """
        if device is None:
            device = "cuda:0"

        device = device.lower()
        if not torch.cuda.is_available():
            device = "cpu"

        self.device = torch.device(device)



from typing import List, Union
from pathlib import Path
import pickle
import numpy as np
from sklearn.svm import SVC
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.decomposition import PCA
from .torch_mlp import TorchMLP
from tactus_data import BodyAngles

AVAILABLE_MODELS = {
    "LSTM": "LSTM",
    "SVC": SVC,
    "TorchMLP": TorchMLP,
    "GradientBoostingClassifier": GradientBoostingClassifier,
}


class Classifier:
    ACTION_LABELS = ['kicking', 'punching', 'pushing', 'neutral']

    def __init__(self, classifier_name: str = None, hyperparams: dict = None) -> None:
        """
        Create the classifier instance with the given name and
        hyperparametres.

        Parameters
        ----------
        classifier_name : str, optional
            name of the classifier. Must be in "SVC", "MLPClassifier",
            "GradientBoostingClassifier", by default None.
        hyperparams : dict, optional
            dictionnary of the classifier hyperparametres,
            by default None.
        """
        self.clf = None
        self.name = None
        self.hyperparams = None

        self.window_size: int = None
        self._angles_to_compute: List[BodyAngles] = None
        self.fps: int = None

        if classifier_name is not None and hyperparams is not None:
            self.name = classifier_name
            self.hyperparams = hyperparams
            self.clf = AVAILABLE_MODELS[classifier_name](**hyperparams)
        self.pca = None

    @classmethod
    def load(cls, model_weights_path: Path) -> "Classifier":
        """
        load model weights from a path.

        Parameters
        ----------
        model_weights_path : Path
            path to the model weights.

        Returns
        -------
        Classifier
            return instance of the classifier to allow chaining like
        Examples
        --------
        `clf = Classifier.load(model_weights_path)`.
        """
        return pickle.load(model_weights_path.open("rb"))

    def save(self, save_path: Path):
        """
        save the model weights inside a pickle file.

        Parameters
        ----------
        save_path : Path
            where to save the model weights.
        """
        save_path.parent.mkdir(parents=True, exist_ok=True)
        pickle.dump(self, save_path.open(mode="wb"))

    def fit_pca(self, X: Union[np.ndarray, List[List]], Y = None, *, min_pca_features: int = 50):
        """
        Fit the model with X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.
        Y :
            ignored
        min_pca_features : int, optional
            minimum number of features after the pca, by default 50.

        Returns
        -------
        PCA : object
            Returns the instance itself.
        """
        n_components = min(min_pca_features, X.shape[1])
        self.pca = PCA(n_components=n_components, whiten=True)
        return self.pca.fit(X)

    def transform(self, X: Union[np.ndarray, List[List]]):
        """
        Apply dimensionality reduction to X.

        X is projected on the first principal components previously extracted
        from a training set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            New data, where `n_samples` is the number of samples
            and `n_features` is the number of features.

        Returns
        -------
        X_new : array-like of shape (n_samples, n_components)
            Projection of X in the first principal components, where
            `n_samples` is the number of samples and `n_components` is
            the number of the components.
        """
        return self.pca.transform(X)

    def fit(self, X, Y):
        """
        fit the classifier to the data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            features.
        Y : array-like of shape (n_samples)
            ground truth.

        Returns
        -------
        self
            return the instance of the classifier.
        """
        return self.clf.fit(X, Y)

    def predict(self, X) -> List[int]:
        """
        predict the action index on a feature set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            features.

        Returns
        -------
        array-like of shape (n_samples)
            the predicted indexes.
        """
        return self.clf.predict(X).tolist()

    def predict_label(self, X) -> List[str]:
        """
        predict the label on a feature set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            features.

        Returns
        -------
        array-like of shape (n_samples)
            the predicted labels.
        """
        pred = self.predict(X)

        labels = [""] * len(pred)
        if hasattr(self, "ACTION_LABELS"):
            labels = [self.ACTION_LABELS[index] for index in pred]
        else:
            # for retro compatibility purposes
            labels = [int_to_label(index) for index in pred]

        return labels

    @property
    def angles_to_compute(self) -> List[BodyAngles]:
        """for compatibility purpose."""
        if hasattr(self, "_angles_to_compute"):
            return self._angles_to_compute
        else:
            DeprecationWarning("This instance has an depreciated attribute ",
                               "angle_to_compute. Use angles_to_compute instead.")
            return self.angle_to_compute

    @angles_to_compute.setter
    def angles_to_compute(self, values: List[BodyAngles]):
        self._angles_to_compute = values


def label_to_int(label: str) -> int:
    """transform a label to its corresponding integer."""
    return Classifier.ACTION_LABELS.index(label)


def int_to_label(index: int) -> str:
    """transform an int to its corresponding label."""
    return Classifier.ACTION_LABELS[index]
